# coding=utf-8
"""
NetMagus Python Library
Copyright (C) 2016 Intelligent Visbility, Inc.
Richard Collins <richardc@intelligentvisibility.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import importlib
import json
import logging
import os
import pickle
import traceback

import netmagus.form
import netmagus.rpc
import netmagus.screen


class NetMagusSession(object):
    """
    This class is a wrapper to handle an interactive session between the
    NetMagus backend and a user script.  It serves as a unifed API for user
    script operations to send/receive data with the NetMagus backend.  It
    also serves as a unified API for the NetMagus backend to call and execute
    user scripts.

    The NetMagus backend will call the module as (ex.):
        ``python -m netmagus --script user.py --token aBcDeF --input-file
        /some/path/to/aBcDeF.json --loglevel 10``

    The "token" is used by NM to desginate all file and RPC interactions tied
    to a given execution.

    This class will be called by the module's __main__.py to:
     - read JSON from the NetMagus backend via the input-file if it exists
     - read in any previous state tied to the token/session
     - import the user's script/module and execute it's run() method
        - run() method must receive a NetMagusSession object as only arg
        - run() method may return Form or (Form, anyobject)
     - receive a Form (and a state object) as a return from the user's
       run() method
     - store any state object and send JSON response back to NetMagus backend


     :meth:`.start` is used to initiate the execution of the user's code
     in the formula.

     The user's code can use various attributes and methods of this session
     object to interact with the NetMagus server and its UI.

     The following attributes exist within each session object:
     * :attr:`nm_input`:  holds the data entered by end-user on the NM ui
     * :attr:`user_state`: holds any state data from previous formula steps

    """

    def __init__(self, token, input_file, loglevel, script):
        """
        The netmagus package's main method will parse the CLI args passed from
        the NetMagus backend and instantiate a new NetMagusSession object used
        to manage the connection and data passing to/from the NetMagus backend.
        
        :param token: a randomized token used to associate with a given
                      formula execution, provided from NM
        :type token: str
        :param input_file: the abs path to an input JSON file from NM backend
        :type input_file: str
        :param loglevel: a value from 0-5 indicating the log level to use
                         for the debugger passed from NM backend
        :type loglevel: int
        :param script: the name of the user python module to import and call
                       the run() entry point method
        :type script: str
        :returns: The next operation to be executed when this one completes
        :rtype: netmagus.netop.NetOpStep
        """
        # the NM "token" used to differntiate every commandPath execution
        self.token = token
        # the JSON file sent from NM back-end as input to this execution
        self.input_file = input_file
        # the logging level set in the NM UI formula admin screen
        self.loglevel = loglevel
        # the name of the Python module in the formula's directory
        self.script = script
        # the session logger setup when self.start() is called
        self.logger = logging.getLogger(__name__)
        # nm_input stores the JSON file input from NetMagus backend
        self.nm_input = None
        # session state data to be used by multiple screen.Screen objects
        # if they need to store/pass persistent data between multiple screens
        self.session_data = None
        self.user_state = self.session_data  # legacy ref for old versions
        # Convenience methods for user script writers
        self.rpc_connect = netmagus.rpc.rpc_connect
        self.rpc_disconnect = netmagus.rpc.rpc_disconnect
        self.rpc_form = netmagus.form.Form
        self.rpc_html = netmagus.rpc.Html
        self.netopstep = netmagus.form.Form  # for legacy <v0.6.0 compatability
        self.form = netmagus.form.Form
        self.textarea = netmagus.form.TextArea
        self.textinput = netmagus.form.TextInput
        self.radiobutton = netmagus.form.RadioButton
        self.passwordinput = netmagus.form.PasswordInput
        self.dropdownmenu = netmagus.form.DropDownMenu
        self.checkbox = netmagus.form.CheckBox

    def rpc_send(self, message):
        """
        Send a message to the NetMagus backend via WAMP to the session's RPC
        target.

        This is a convenience method to handle the token manipulation for the
        session. See :meth:`netmagus.rpc.rpc_send` method for full args
        """
        netmagus.rpc.rpc_send(self.token, message)

    def rpc_receive(self):
        """
        Retrieve a message via WAMP from the session's RPC target on the
        NetMagus backend service.

        This is a convenience method to handle the token manipulation for the
        session.  See :meth:`netmagus.rpc.rpc_receive` method for full args

        :returns: Any response message currently available at the RPC target
            for this session
        """
        return netmagus.rpc.rpc_receive(self.token)

    def rpc_form_query(self, message, **kwargs):
        """
        Send a message to the NetMagus backend via WAMP to the session's RPC
        target, then poll the target for response data.

        This is a convenience method to handle token manipulation for the
        session.  See :meth:`netmagus.rpc.rpc_form_query` for a full set of args

        :param message: an rpc.Message object to be sent to NetMagus backend
        :param kwargs: see rpc.rpc_form_query for full arg list
        :returns: the response messsage at the target RPC target for
            this session

        """
        return netmagus.rpc.rpc_form_query(self.token, message, **kwargs)

    def rpc_html_clear(self):
        """
        Send a message to the NM backend to clear the UI's HTML pop-up
        display area.
        """
        self.rpc_send(netmagus.rpc.Html(append=False))

    def start(self):
        """
        Method to start the execution of user python module for the session

        :returns:  a JSON result file and a Shelve to store a user state object
        """
        self.logger.setLevel(self.loglevel)
        self.__read_response_file()
        self.__read_state_file()
        formula_return = self.__run_user_script()
        self.__write_response_file(formula_return)
        self.__write_state_file(self.user_state)

    def __run_user_script(self):
        """
        Used by the start method to import and execute the user's Python module
        containing the formula logic

        :return: the netmagus.form.Form object returned by the user's module.
        """
        try:
            self.logger.debug('Attempting to import user module: '
                              '{}'.format(self.script))
            user_module = importlib.import_module(self.script)
            self.logger.debug('User module imported: {}'.format(user_module))
            self.logger.debug('Attempting to execute user module run() method')
            formula_return = user_module.run(self)
            if isinstance(formula_return, netmagus.form.Form):
                self.logger.debug('User module returned a form only')
                self.__fix_formcomponent_indexes(formula_return)
                return formula_return
            # allow for old formula format <v0.6.0 with tuple
            elif isinstance(formula_return, tuple):
                if isinstance(formula_return[0], netmagus.form.Form):
                    self.__write_state_file(formula_return[1])
                    return formula_return[0]
            else:
                raise TypeError('Formula files should return a '
                                'netmagus.form.Form object')
        except ImportError:
            raise ImportError('Unable to load user module: '
                              '{}'.format(self.script))
        except (Exception, NameError, IOError) as ex:
            self.logger.critical(
                'Error calling run() method defined in the target file: '
                '{}'.format(self.script))
            tb = traceback.format_exc()
            logging.exception(ex)
            htmlextrainfo = {
                'html': {
                    'printToUser': True,
                    'outputType': 'html',
                    'title': 'ERROR IN FORMULA',
                    'data': '<h3>This formula has encountered a critical error '
                            'and can not continue.  Please review the '
                            'traceback info and/or contact support for '
                            'assistance.</h3><br><br>Traceback info was: '
                            '<pre>{}</pre>'.format(tb)
                }
            }
            next_step = netmagus.form.Form(currentStep=999, dataValid=False,
                                           extraInfo=htmlextrainfo,
                                           disableBackButton=True,
                                           finalStep=True)
            return next_step

    def __read_response_file(self):
        """
        Read in the JSON response file from the NetMagus back-end.  These files
        are generated each time the NetMagus backend executes a commandPath to
        launch a task defined in a formula.  Examples would be when a user
        presses the SUBMIT button in the UI, a JSON file is generated by
        NetMagus and stored in a temp file to pass data to the Formula.
        """
        self.logger.debug('Reading JSON data from NetMagus request')
        try:
            with open(self.input_file) as data_file:
                self.nm_input = json.load(data_file)
            os.remove(self.input_file)  # remove file after reading it
        except IOError:
            self.logger.warn('Unable to access input JSON file {}'.format(
                self.input_file))
        except TypeError:
            self.logger.error('Unable to decode JSON data in {}'.format(
                self.input_file))

    def __read_state_file(self):
        """
        This method will retrieve any previous state data saved during this
        formula execution and store it internal as self.user_state where it
        can be used throughout the formula execution and serve as a target
        for persistent data storage throughout the formula's multiple
        execution steps.
        """
        state_file = self.input_file + '_State'
        try:
            with open(state_file, 'rb') as picklefile:
                self.user_state = pickle.load(picklefile)
                self.logger.debug('Formula state retrieved from '
                                  '{}'.format(state_file))
        except pickle.UnpicklingError:
            self.logger.info('The state file {} exists but does not '
                             'contain previous state from this formula '
                             'execution.  Setting state to '
                             'NONE'.format(state_file))
            self.user_state = None
        except IOError:
            self.logger.info('No _State file found from previous formula '
                             'execution steps. Setting state to NONE.')
            self.user_state = None

    def __write_response_file(self, response):
        """
        Store the returned Form into a Response file for NetMagus to read and
        process for execution of the next formula step
        """
        response_file = self.input_file + 'Response'
        self.logger.debug(
            'Target output JSON file will be: {}'.format(response_file))
        try:
            with open(response_file, 'w') as outfile:
                outfile.write(str(response))
        except IOError:
            self.logger.error('Unable to open target JSON Response file: '
                              '{}'.format(response_file))
            raise

    def __write_state_file(self, stateobject):
        """
        store the returned state object into a file for future operation
        steps to retrieve.  Formula creators can store an object in the
        sessions user_state attribute to have anyobject saved to disk and
        passed to the next formula execution step where it will be retrieved
        and stored in NetMagusSession.user_state for use by other formula code
        """
        state_file = self.input_file + '_State'
        self.logger.debug(
            'Target output shelve file will be: {}'.format(state_file))
        try:
            with open(state_file, 'wb') as picklefile:
                pickle.dump(stateobject, picklefile, protocol=-1)
                self.logger.debug('Formula state stored in '
                                  '{}'.format(state_file))
        except pickle.PickleError:
            self.logger.error('Error pickling state object')
            raise
        except IOError:
            self.logger.error('Unable to open target state file: '
                              '{}'.format(state_file))
            raise

    @staticmethod
    def __fix_formcomponent_indexes(form_obj):
        """
        This method is a temporary fix to append an index attribute to each
        form component to be sent to NetMagus as JSON.  Eventually this will
        be done in the NetMagus back-end upon receipt according to the order
        of the list of form controls.  For now these are being added here in
        the same fashion before being sent to the NetMagus back-end.
        :param form_obj: a netmagus.NetOpStep object to be serialized and
        sent to NetMagus
        """
        # TODO: REMOVE INDEX INCREMENTING ONCE PROBERT100 FIXES NM
        index_counter = 0
        for item in form_obj.form:
            setattr(item, 'index', index_counter)
            index_counter += 1

    def display_screen(self, screen):
        """
        Render a :class:`netmagus.screen.ScreenBase` object to the NetMagus UI
        for this session.

        You may optionally also pass in a callable that returns a
        :class:`netmagus.screen.ScreenBase` object.
        The passed function will be called and if it returns a valid instance
        it will be used.

        :param screen: the screen to activate in the UI
        :type screen: netmagus.screen.ScreenBase
        :return: a final form to be shown at session completion
        :rtype: netmagus.form.Form
        """
        # validate type of screen
        if callable(screen):
            self.logger.debug('dispaly_screen executing callable {} to generate'
                              ' Screen'.format(screen))
            screen = screen()
        if not isinstance(screen, netmagus.screen.ScreenBase):
            raise TypeError('session.display argument is not '
                            'a valid Screen object')
        self.logger.debug('displaying screen {}'.format(screen))
        # set the screen's session attribute to this session
        screen.session = self
        try:
            # reset flag each time screen is displayed
            screen.input_valid = False
            while not screen.input_valid:
                self.logger.debug('sending form to UI')
                response = screen.session.rpc_form_query(screen.form)
                screen.user_input = response[
                    'wellFormatedInput']  # store user input
                # don't log actual user input since it may contain sensative
                # data that could be viewd in debug output or formula history
                # in the UI
                self.logger.debug('{} user input received with length {}'
                                  ''.format(screen.__class__.__name__,
                                            len(screen.user_input)
                                            )
                                  )
                screen.button_pressed = response.get('pressedButton')
                # handle the button press to apply next/back/cancel logic
                screen.process_button()
            # clear HTML pop-up screen once valid data received
            if screen.clear_html_popup:
                screen.session.rpc_html_clear()
            if screen.next_screen:
                self.logger.debug('Data validation passed, '
                                  'processing next screen')
                self.display_screen(screen.next_screen)
            else:
                self.logger.debug('No next_screen set, returning')
        except netmagus.screen.CancelButtonPressed:
            self.rpc_html_clear()
            screen.handle_cancel_button()
            raise
        except netmagus.screen.BackButtonPressed:
            screen.handle_back_button()
            self.display_screen(screen.back_screen)



# coding=utf-8
"""
NetMagus Python Library
Copyright (C) 2016 Intelligent Visbility, Inc.
Richard Collins <richardc@intelligentvisibility.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import abc
import logging
import sys

if sys.version_info >= (3, 4):
    ABC = abc.ABC
else:
    ABC = abc.ABCMeta(str('ABC'), (), {})

import netmagus.form
import netmagus.session

logger = logging.getLogger(__name__)


class CancelButtonPressed(Exception):
    pass


class BackButtonPressed(Exception):
    pass


class ScreenBase(ABC):
    """
    Screen class represents a typical usage pattern for NetMagus form wizards.

    A Form is presented for user input, user data is entered, then data is
    validated.  If the data is invalid, show a pop-up HTML dialog and allow
    user to correct the entry.  Continue to loop until valid data is provided.
    If valid data received then transform response to correct data types, or
    store the input, or do additional workneeded for business logic and return.

    A "next" button with provided name is displayed to execute data validation
    process.

    A "cancel" button is displayed to allow the user to exit on this screen.

    Optionally present the user an additional "back" button to jump back to a
    previous Screen in the formula.

    This is an abstract base class and requires sublass implementations to obey
    the contract for the defined abstractmethods.  Subclasses may over-ride or
    call super() upon the concrete methods (such as __init__ and process_button)
    as desired.

    Example usage:

    Note: self.session is set when NetMagusSession.display_ScreenBase() is
    called to render a ScreenBase() in the UI ::

        class MinScreen(ScreenBase):

            def generate_form(self):
                inputfield1 = form.TextInput(label='field1')
                return self.form = form.Form(name='Form1',
                                      description='Some Description',
                                      form=[inputfield1])

            def handle_back_button(self):
                # perform any clean up here
                pass

            def handle_cancel_button(self):
                # perform any clean up here
                pass

            def process_user_input(self):
                # perform any input transformation or work here
                pass

            def validate_user_input(self):
                # perform data validation here
                return True

            def return_error_message(self):
                # perform any work and then send error to the user when
                # data is invalid
                rpc.rpc_send(rpc.Html(title='Error', data='Data'))

            def return_screen_result(self):
                # perform any work and store the screen results if any
                self.session.session_data['data1'] = self.user_input['field1']

    """

    def __init__(self, session=None, clear_html_popup=True):
        """

        :param session: the NetMagusSession using this form
        :type session: netmagus.session.NetMagusSession
        """
        # the NetMagusSession that is using this screen
        self.session = session
        # the wellFormatedInput returned from the form
        self.user_input = None
        # flag to indicate that user input passed validation
        self.input_valid = False
        # used to track which button was pressed on the screen
        self.button_pressed = None
        # pointer to the next netmagus.screen.ScreenBase() instance to show
        self.next_screen = None
        # pointer to the netmagus.screen.ScreenBase() object if back is pressed
        self.back_screen = None
        # flag to control if this screen should clear the HTML pop-up area after
        # data validation passes
        self.clear_html_popup = clear_html_popup

    @property
    def form(self):
        """
        property to return the form generated by the user-provided
        self.generate_form() method
        """
        return self._render_form()

    def _render_form(self):
        """
        Calls the user's generate_form method and then does any desired clean
        up of the form attributes:

        * Auto-disables the back button if no back_screen is set

        """
        user_form = self.generate_form()
        if not self.back_screen:
            user_form.disableBackButton = True
        return user_form

    def _process_user_input(self):
        """
        Used to track internally if the user overridden data validation
        function has passed and trigger formula to continue, reprompt for
        valid input, or error out

        """
        if self.validate_user_input():
            logger.debug('User input passed data validation')
            self.input_valid = True
            self.process_user_input()
        else:
            logger.debug('User input FAILED data validation')
            self.input_valid = False
            self.return_error_message()

    def process_button(self):
        """
        Provides default logic to handle the button pressed by the user
        on this screen.

        - if button pressed was "cancel" execute formula clean-up logic and
            raise an exception to be handled by the caller
        - if button pressed was "back" go to previous screen
        - if button pressed was "next" validate and return data

        This may be over-ridden to provide custom handling as needed.

        """
        logger.debug('UI button pressed: {}'.format(self.button_pressed))
        if self.button_pressed == 'cancel':
            raise CancelButtonPressed
        elif self.button_pressed == 'back':
            raise BackButtonPressed
        elif self.button_pressed.lower() == 'next':
            self._process_user_input()
        else:
            raise NotImplementedError('Handler for button {} not defined'
                                      .format(self.button_pressed))

    @abc.abstractmethod
    def validate_user_input(self):  # type: () -> bool
        """
        This method should be over-ridden by sublass with screen-specific
        logic to validate the value in :attr:`self.user_input`

        This method should simply indicate if the data entered by the user is
        adequate to proceed to call process_user_input and then proceed to
        display the next screen referenced in :attr:`self.next_screen`

        :return: True/False if data is Valid/Invalid
        :rtype: bool
        """
        raise NotImplementedError

    @abc.abstractmethod
    def process_user_input(self):
        """
        Overide this method when subclassing to process any user input,
        store the results for access by other screens if necessary, and set
        any variable Screen.next_screen values to steer formula execution.

        Persistent data for a set of Screens may be stored in
        :attr:`netmagus.session.NetMagusSession.session_data`
        (which is accessible to all Screen objects in the same session).  Or you may opt to store it in a global or other
        namespace objet usable by all Screens.

        This is used primarily in cases where formula may allow users to go
        back to previous screens or fork in different decision trees.  Screens
        at the end of a decision tree of screens may need to access prior screen
        data to perform operations upon the data from all or a subset of screens

        ex - read from self.user_input['IP Address'] and then generate an
        access-list entry in proper format and do
        self.screen_data='deny ip host 1.2.3.4 any' so that other screens can
        retrieve that ACL entry to be used in some other step of a process.

        """
        return NotImplementedError

    @abc.abstractmethod
    def return_error_message(self):
        """
        Overide this method when sublcassing to send any desired HTML
        messages to the UI when user input data validation fails

        :attr:`self.session.rpc_html` instances can be sent via the
        :meth:`self.session.rpc_send` method to appear in the HTML pop-up area
        of the UI

        """
        raise NotImplementedError

    @abc.abstractmethod
    def handle_cancel_button(self):
        """
        Overide this method when sublcassing to take the desired actions when a
        user hits the cancel button during a formula execution step.

        This method should perform any required clean-up work and/or update any
        other forms that are displayed upon cancellation.

        After this is called by
        :meth:`netmagus.session.NetMagusSession.display_screen` a
        :exc:`netmagus.screen.CancelButtonPressed` exception will be raised to
        allow the caller to display an appropriate :class:`netmagus.form.Form`
        when the formula exits.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def handle_back_button(self):
        """
        Overide this method when subclassing to take desired action when a user
        presses the back button in a formula execution step.

        Use this method to perform any necessary clean-up work when going back
        to a previous screen.

        After this method is called by
        :meth:`netmagus.session.NetMagusSession.display_screen`, it will then
        auto-display the screen referenced in :attr:`self.back_screen`

        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_form(self):  # type: () -> netmagus.form.Form
        """
        Overide this method when sublcassing to generate the desired NetMagus
        form to be seen in the UI.

        :return: the desired form with relevant form controls for user
            interaction
        :rtype: netmagus.form.Form

        """
        raise NotImplementedError



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals


import json

import pytest

from netmagus.form import CheckBox, DropDownMenu, PasswordInput, RadioButton, TextArea, TextInput
from netmagus.netop import NetOpStep


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit,
# PyAttributeOutsideInit
# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit,
# PyAttributeOutsideInit
# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit,
# PyAttributeOutsideInit
# noinspection PyAttributeOutsideInit
class TestNetMagusNetOp(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.sample_netop = json.dumps(
                {
                    'id': 0, 'name': 'My Network Operation',
                    'commandPath': 'python -u /tmp/myfile.py',
                    'description': 'My Description',
                    'currentStep': 2,
                    'form': [
                        {
                            "component": "textArea", "editable": True,
                            "index": 0, "label": "Label 1",
                            "description": "Description 1",
                            "placeholder": "Placeholder 1", "options": [],
                            "required": True, "validation": "/.*/",
                            "value": None
                        }, {
                            "component": "textInput", "editable": True,
                            "index": 0, "label": "Label 2",
                            "description": "Description 2",
                            "placeholder": "Placeholder 2", "options": [],
                            "required": True, "validation": "[number]",
                            "value": None
                        }, {
                            "component": "checkbox", "editable": True,
                            "index": 0, "label": "Label 3",
                            "description": "Description 3", "placeholder": "",
                            "options": ["Option1", "Option2"], "required": True,
                            "validation": "/.*/",
                            "value": "Option1"
                        }, {
                            "component": "radio", "editable": True, "index": 0,
                            "label": "Label 4",
                            "description": "Description 4", "placeholder": "",
                            "options": ["Option1", "Option2"], "required": True,
                            "validation": "/.*/",
                            "value": "Option1"
                        }, {
                            "component": "select", "editable": True, "index": 0,
                            "label": "Label 5",
                            "description": "Description 5", "placeholder": "",
                            "options": ["Option1", "Option2"], "required": True,
                            "validation": "/.*/",
                            "value": "Option1"
                        }, {
                            "component": "password", "editable": True,
                            "index": 0, "label": "Label 6",
                            "description": "Description 6",
                            "placeholder": "Placeholder 6", "options": [],
                            "required": True, "validation": "/.*/",
                            "value": None
                        },
                    ],
                    'extraInfo': {
                        "hiddenData1": "Just Some data1",
                        "hiddenData2": {
                            "printToUser": True,
                            "outputType": "html",
                            "data": "Just Some data2"
                        },
                        "image1": {
                            "printToUser": True, "outputType": "img",
                            "data": "img/image1.png"
                        }
                    },
                    'dataValid': False,
                    'buttonName': 'Next',
                    'saveRecord': True,
                    'dynamic': True,
                    'autoProceed': True,
                    'disableBackButton': True,
                    'finalStep': True
                }, sort_keys=True)

        self.extrainfo = {
            'hiddenData1': 'Just Some data1',
            'hiddenData2': {
                'data': "Just Some data2", 'outputType': 'html',
                'printToUser': True
            },
            'image1': {
                'printToUser': True, 'outputType': 'img',
                'data': 'img/image1.png'
            }
        }

        self.textarea = TextArea(editable=True, label='Label 1',
                                 description='Description 1',
                                 placeholder='Placeholder 1', required=True)
        self.textinput = TextInput(editable=True, label='Label 2',
                                   description='Description 2',
                                   placeholder='Placeholder 2', required=True,
                                   validation='[number]')
        self.checkbox = CheckBox(editable=True, label='Label 3',
                                 description='Description 3',
                                 options=['Option1', 'Option2'], required=True,
                                 value='Option1')
        self.radiobutton = RadioButton(editable=True, label='Label 4',
                                       description='Description 4',
                                       options=['Option1', 'Option2'],
                                       required=True, value='Option1')
        self.dropdown = DropDownMenu(editable=True, label='Label 5',
                                     description='Description 5',
                                     options=['Option1', 'Option2'],
                                     required=True, value='Option1')
        self.password = PasswordInput(editable=True, label='Label 6',
                                      description='Description 6',
                                      placeholder='Placeholder 6',
                                      required=True)
        self.form = [self.textarea, self.textinput, self.checkbox,
                     self.radiobutton, self.dropdown, self.password]

        self.netopstep = NetOpStep(name='My Network Operation',
                                   commandPath='python -u /tmp/myfile.py',
                                   description='My Description', form=self.form,
                                   extraInfo=self.extrainfo, currentStep=2,
                                   dataValid=False,
                                   buttonName='Next', saveRecord=True,
                                   dynamic=True, autoProceed=True,
                                   disableBackButton=True, finalStep=True)
        return self

    def test_netop_obj(self, components):
        assert components.sample_netop == repr(components.netopstep)
        assert json.dumps(components.netopstep.as_dict,
                          sort_keys=True) == components.sample_netop
        assert repr(components.netopstep) == json.dumps(
                components.netopstep.as_dict, sort_keys=True)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals


import json

import pytest

from netmagus.form import CheckBox, DropDownMenu, PasswordInput, RadioButton, TextArea, TextInput


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusFormTextArea(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.ta = TextArea(editable=True, label='My Label', description='My Description',
                           placeholder='My Placeholder', required=True)
        # sample form element as created in the NetMagus Form Designer UI
        self.ta_fd = json.loads('{"component": "textArea", "editable": true, "index": 0, "label": "My '
                                'Label", '
                                '"description": "My Description", "placeholder": "My Placeholder", '
                                '"options": [], '
                                '"required": true, "validation": "/.*/", "value": null}')
        return self

    def test_textarea_obj(self, components):
        assert components.ta.__dict__ == components.ta_fd

    def test_textarea_json(self, components):
        assert repr(components.ta) == json.dumps(components.ta_fd, sort_keys=True)
        assert json.dumps(components.ta.as_dict, sort_keys=True) == json.dumps(components.ta_fd, sort_keys=True)


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusFormTextInput(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.ti = TextInput(editable=True, label='My Label',
                            description='My Description',
                            placeholder='My Placeholder', required=True,
                            validation='[number]')
        # sample form element as created in the NetMagus Form Designer UI
        self.ti_fd = json.loads(
                '{"component":"textInput","editable":true,"index":0,'
                '"label":"My '
                'Label",'
                '"description":"My Description","placeholder":"My Placeholder",'
                '"options":[],'
                '"required":true,"validation":"[number]", "value": null}')
        return self

    def test_textinput_json(self, components):
        assert repr(components.ti) == json.dumps(components.ti_fd,
                                                 sort_keys=True)

    def test_textinput_obj(self, components):
        assert components.ti.__dict__ == components.ti_fd
        assert json.dumps(components.ti.as_dict, sort_keys=True) == json.dumps(
                components.ti_fd, sort_keys=True)


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusFormCheckbox(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.cb = CheckBox(editable=True, label='My Label',
                           description='My Description',
                           options=['Option 1', 'Option 2'], required=True,
                           value='test')

        self.cb_fd = json.loads(
                '{"component":"checkbox","editable":true,"index":0,"label":"My '
                'Label",'
                '"description":"My Description","placeholder":"","options":['
                '"Option 1","Option 2"],"required":true,"validation":"/.*/", '
                '"value":"test"}')
        return self

    def test_checkbox_json(self, components):
        assert repr(components.cb) == json.dumps(components.cb_fd,
                                                 sort_keys=True)

    def test_checkbox_obj(self, components):
        assert components.cb.__dict__ == components.cb_fd
        assert json.dumps(components.cb.as_dict, sort_keys=True) == json.dumps(
                components.cb_fd, sort_keys=True)


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusRadioButton(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.rb = RadioButton(editable=True, label='My Label',
                              description='My Description',
                              options=['Option 1', 'Option 2'], value='test')

        self.rb_fd = json.loads(
                '{"component":"radio","editable":true,"index":0,"label":"My '
                'Label",'
                '"description":"My Description","placeholder":"","options":['
                '"Option 1","Option 2"],"required":false,"validation":"/.*/", '
                '"value":"test"}')
        return self

    def test_checkbox_json(self, components):
        assert repr(components.rb) == json.dumps(components.rb_fd,
                                                 sort_keys=True)

    def test_checkbox_obj(self, components):
        assert components.rb.__dict__ == components.rb_fd
        assert json.dumps(components.rb.as_dict, sort_keys=True) == json.dumps(
                components.rb_fd, sort_keys=True)


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusDropDownMenu(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.dd = DropDownMenu(editable=True, label='My Label',
                               description='My Description',
                               options=['Option 1', 'Option 2'], value='test')

        self.dd_fd = json.loads('{"component":"select","description":"My '
                                'Description","editable":true,"index":0,'
                                '"label":"My Label","placeholder":"",'
                                '"options":["Option 1","Option 2"],'
                                '"required":false,"validation":"/.*/",'
                                '"value":"test"}')
        return self

    def test_dropdown_json(self, components):
        assert repr(components.dd) == json.dumps(components.dd_fd,
                                                 sort_keys=True)
        assert json.dumps(components.dd.as_dict, sort_keys=True) == json.dumps(
                components.dd_fd, sort_keys=True)

    def test_dropdown_obj(self, components):
        assert components.dd.__dict__ == components.dd_fd


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit
class TestNetMagusPasswordInput(object):
    @pytest.fixture(scope='class')
    def components(self):
        self.pi = PasswordInput(editable=True, label='My Label',
                                description='My Description',
                                placeholder='My Placeholder', required=True)

        self.pi_fd = json.loads(
                '{"component":"password","editable":true,"index":0,"label":"My '
                'Label",'
                '"description":"My Description","placeholder":"My Placeholder",'
                '"options":[],'
                '"required":true,"validation":"/.*/", "value": null}')
        return self

    def test_password_json(self, components):
        assert repr(components.pi) == json.dumps(components.pi_fd,
                                                 sort_keys=True)
        assert json.dumps(components.pi.as_dict, sort_keys=True) == json.dumps(
                components.pi_fd, sort_keys=True)

    def test_password_obj(self, components):
        assert components.pi.__dict__ == components.pi_fd



# coding=utf-8
__author__ = 'Richard Collins'
__email__ = 'richardc@intelligentvisibility.com'
__version__ = '0.7.2'
__all__ = ['form', 'rpc', 'screen', 'session']
from .form import Form, TextInput, TextArea, CheckBox, DropDownMenu, \
    PasswordInput, RadioButton
from .rpc import Html, rpc_connect, rpc_disconnect, rpc_form_query, rpc_send, \
    rpc_receive
from .screen import ScreenBase, CancelButtonPressed, BackButtonPressed
from .session import NetMagusSession



# coding=utf-8
"""
Crossbar should not be running before running this test suite.  fixtures will
start/stop it as needed for various tests.
"""

from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
import logging
import os
import subprocess
from os import path
from time import sleep, time

import autobahn_sync
import pytest
from autobahn_sync import AutobahnSync
from autobahn_sync.exceptions import ApplicationError, ConnectionRefusedError

from netmagus.form import CheckBox, DropDownMenu, RadioButton, TextArea, \
    TextInput, Form
from netmagus.rpc import Html, Message, RpcCallTimeout, rpc_connect, \
    rpc_disconnect, rpc_form_query, rpc_send


@pytest.fixture(scope='session')
def logger():
    """
    logger fixture used by all tests
    """
    return logging.getLogger(__name__)


@pytest.fixture(scope='module')
def components(request):
    """
    mock data components used for method validation
    """
    request.extrainfo = {
        "hiddenData1": "Just Some data1", "hiddenData2": {
            "printToUser": True, "outputType": "html", "data": "Just Some data2"
        }, "image1": {
            "printToUser": True, "outputType": "img", "data": "img/image1.png"
        }
    }

    request.sample_form_message_dict = {
        'saveRecord': True, 'name': 'Test Form',
        'description': 'Test Description', 'buttonName': 'Test Button',
        'extraInfo': request.extrainfo, 'disableBackButton': True,
        'autoProceed': True, 'currentStep': 2,
        'finalStep': False, 'dataValid': False,
        'commandPath': "", 'dynamic': False, 'id': 0,
        'form': [{
            "component": "textArea", "editable": True, "index": 0,
            "label": "Label 1",
            "description": "Description 1", "placeholder": "Placeholder 1",
            "options": [], "required": True,
            "validation": "/.*/", "value": None
        }, {
            "component": "textInput", "editable": True, "index": 0,
            "label": "Label 2",
            "description": "Description 2", "placeholder": "Placeholder 2",
            "options": [], "required": True,
            "validation": "[number]", "value": None
        }, {
            "component": "checkbox", "editable": True, "index": 0,
            "label": "Label 3",
            "description": "Description 3", "placeholder": "",
            "options": ["Option1", "Option2"], "required": True,
            "validation": "/.*/", "value": "Option1"
        }, {
            "component": "radio", "editable": True, "index": 0,
            "label": "Label 4",
            "description": "Description 4", "placeholder": "",
            "options": ["Option1", "Option2"], "required": True,
            "validation": "/.*/", "value": "Option1"
        }, {
            "component": "select", "editable": True, "index": 0,
            "label": "Label 5",
            "description": "Description 5", "placeholder": "",
            "options": ["Option1", "Option2"], "required": True,
            "validation": "/.*/", "value": "Option1"
        }]
    }

    request.sample_form_json = json.dumps(request.sample_form_message_dict, sort_keys=True)

    request.textarea = TextArea(editable=True, label='Label 1', description='Description 1',
                                placeholder='Placeholder 1', required=True)
    request.textinput = TextInput(editable=True, label='Label 2', description='Description 2',
                                  placeholder='Placeholder 2', required=True, validation='[number]')
    request.checkbox = CheckBox(editable=True, label='Label 3', description='Description 3',
                                options=['Option1', 'Option2'], required=True, value='Option1')
    request.radiobutton = RadioButton(editable=True, label='Label 4', description='Description 4',
                                      options=['Option1', 'Option2'], required=True, value='Option1')
    request.dropdown = DropDownMenu(editable=True, label='Label 5', description='Description 5',
                                    options=['Option1', 'Option2'], required=True, value='Option1')
    request.form = [request.textarea, request.textinput, request.checkbox, request.radiobutton, request.dropdown]

    request.rpcform = Form(name='Test Form', description='Test Description', buttonName='Test Button',
                           extraInfo=request.extrainfo, disableBackButton=True, dataValid=False,
                           finalStep=False, currentStep=2, autoProceed=True,
                           form=[request.textarea, request.textinput,
                                 request.checkbox, request.radiobutton,
                                 request.dropdown], )

    request.emptyform = Form()

    request.sample_html_message_dict = {
        'printToUser': True, 'outputType': 'html', 'append': True, 'title': 'Test Title',
        'data': '<p>Test html block</p>'
    }

    request.sample_html_json = json.dumps(request.sample_html_message_dict, sort_keys=True)

    request.htmlform = Html(title='Test Title', data='<p>Test html block</p>')

    request.sample_rpcmessage_html_json = json.dumps({
        'messageType': 'html', 'message': request.sample_html_message_dict
    }, sort_keys=True)

    request.sample_rpcmessage_form_json = json.dumps({
        'messageType': 'netop', 'message': request.sample_form_message_dict
    }, sort_keys=True)

    request.rpcmessage_html = Message(request.htmlform)
    request.rpcmessage_form = Message(request.rpcform)
    request.rpcmessage_emptyform = Message(request.emptyform)

    return request


class TestNetMagusRPCForm(object):
    """
    This class of tests verify the rpc.Form and rpc.Html data objects to ensure
    their JSON representations are as expected
    """

    def test_rpcform_json(self, components):
        assert json.dumps(components.rpcform.as_dict,
                          sort_keys=True) == components.sample_form_json
        assert repr(components.rpcform) == \
               json.dumps(components.rpcform.as_dict, sort_keys=True)

    def test_htmlform_json(self, components):
        assert json.dumps(components.htmlform.as_dict,
                          sort_keys=True) == components.sample_html_json
        assert repr(components.htmlform) == str(components.htmlform.as_dict)

    def test_rpcmessages_json(self, components):
        assert json.dumps(components.rpcmessage_html.as_dict,
                          sort_keys=True) == \
               components.sample_rpcmessage_html_json
        assert json.dumps(components.rpcmessage_form.as_dict,
                          sort_keys=True) == \
               components.sample_rpcmessage_form_json
        assert repr(components.rpcmessage_form) == str(
                components.rpcmessage_form.as_dict)


class TestRPCMethodsWithoutCrossbar(object):
    """
    This class is for tests where we do NOT want the Crossbar.io backend running
    to respond to any tests
    """

    def test_rpc_connect_failure(self):
        """
        Ensure that if Crossbar.io is not running, a proper exception is raised
        """
        with pytest.raises(ConnectionRefusedError):
            rpc_connect(url='ws://127.0.0.1:8088/ws', realm='netmagus')


# noinspection PyAttributeOutsideInit,PyAttributeOutsideInit,
# PyAttributeOutsideInit
# noinspection PyAttributeOutsideInit
class TestRPCMethodsWithCrossbar(object):
    """
    This class should hold any tests that require Crossbar.io to run
    """

    @pytest.fixture(scope='module')
    def crossbar(self, request):
        """ make sure a crossbar router is reachable on default url/realm,
        if not start one up.  stop it upon teardown"""
        crossbar_conf_dir = path.abspath(path.dirname(__file__)) + '/.crossbar'
        print('crossbar setup')
        print('current path is: '.format(os.environ['PATH']))
        running = False

        self.userdata = None
        self.received_timestamp = None
        self.ack_timestamp = 0
        self.ack_userdata = None
        # start an autobahn_sync app runnign and register a test RPC method
        # that will receive a message and another
        # to return it back to us
        test_app = AutobahnSync()

        @test_app.register("com.intelligentvisibility.a1b2c3.browser.sendform")
        def sendform(msg):
            """
            Register a fake rpc method similar to that used by NetMagus for
            clients to send data to be rendered.
            """
            self.userdata = msg
            self.received_timestamp = time()
            return "ok"

        @test_app.register(
                "com.intelligentvisibility.a1b2c3.browser.getformresponse")
        def getformresponse():
            """
            Register a fake RPC method to simualte NetMagus responding to a
            client request for data after a user has entered it in the NetMagus
            UI.  only return data if it's been > 2 seconds since we received it
            to simulate processing time by user.  2 second time point allows to
            test cases where we do not wait long enough and get not data due to
            timeout, cases where we wait till end of our client-side timeout and
            raise exception on client side to handle or retry, and cases where
            we block forever with no timeout and the data finally returns at 2s.
            """
            if time() >= self.received_timestamp + 1:
                return self.userdata
            else:
                return False

        @test_app.register(
                "com.intelligentvisibility.a1b2c3.browser.displaymessage")
        def displaymessage(msg):
            """
            Register a fake RPC method to receive Html update messages
            """
            self.userdata = msg
            self.received_timestamp = time()
            return "ok"

        @test_app.register("com.intelligentvisibility.d4e5f6.browser.sendform")
        def no_valid_ui_target(test_app):
            """
            Simulate case where NM backend is online with valid reg, but
            browser UI has not registered its corresponding method yet. In this
            case today NM passes on the exception to us and we should see it.
            """
            return autobahn_sync.call(
                    "com.intelligentvisibility.d4e5f6.ui.browser.sendform")

        @test_app.register("com.intelligentvisibility.123456.browser.sendform")
        def sendform_ack_test(msg):
            """
            Register a fake rpc method similar to that used by NetMagus for
            clients to send data to be rendered.  This endpoint should simulate
            case where backend is not responding for a period to the request and
            will induce retry behavior in the rpc_send method
            """
            self.userdata = msg
            self.received_timestamp = time()
            if self.ack_timestamp == 0:
                self.ack_timestamp = time()
            # only return ACK if it's been 1 second since the request to
            # this method
            if time() >= self.ack_timestamp + 1:
                self.ack_timestamp = 0
                return "ok"

        @test_app.register(
                "com.intelligentvisibility.123456.browser.getformresponse")
        def getformresponse():
            """
            Register a fake RPC method to simualte NetMagus responding to a
            client request for data after a user has entered it in the NetMagus
            UI.  only return data if it's been > 2 seconds since we received it
            to simulate processing time by user.  2 second time point allows to
            test cases where we do not wait long enough and get not data due to
            timeout, cases where we wait till end of our client-side timeout and
            raise exception on client side to handle or retry, and cases where
            we block forever with no timeout and the data finally returns at 2s.
            """
            if time() >= self.received_timestamp + 1:
                return self.userdata
            else:
                return False


        # start up the crossbar service
        for _ in range(20):
            sleep(1)
            try:
                test_app.run(url='ws://127.0.0.1:8088/ws', realm='netmagus')
            except ConnectionRefusedError:
                subprocess.Popen(["crossbar", "start", "--cbdir",
                                  crossbar_conf_dir])
                continue
            else:
                running = True
                break
        if not running:
            raise RuntimeError("Couldn't connect to crossbar router")

        def finalizer():
            """
            tear down Crossbar.io when we are done
            """
            print('crossbar teardown')
            rpc_disconnect()
            p = subprocess.Popen(["crossbar", "stop", "--cbdir",
                                  crossbar_conf_dir])
            p.wait()

        request.addfinalizer(finalizer)
        return self

    def test_rpc_connect_success(self, crossbar):
        """
        verify that when crossbar is online, we can connect to it
        """
        rpc_connect(url='ws://127.0.0.1:8088/ws', realm='netmagus')

    def test_formquery_failure_message_html(self, crossbar, components):
        """
        when crossbar.io is not online, verify we get a proper exception raised
        """
        with pytest.raises(TypeError):
            rpc_form_query('a1b2c3', components.htmlform, poll=.5, timeout=5)

    def test_formquery_success_message_form(self, crossbar, components):
        """
        send rpc.Form to mock RPC method via Crossbar, wait for up to 5s, mock
        method will return back the data sent to it after 2s delay
        """
        assert rpc_form_query('a1b2c3', components.rpcform, poll=.5,
                              timeout=5) == components.rpcmessage_form.as_dict

    def test_formquery_success_emptyform(self, crossbar, components):
        """
        send rpc.Form to mock RPC method via Crossbar, wait for up to 5s, mock
        method will return back the data sent to it after 2s delay
        """
        assert rpc_form_query('a1b2c3', components.emptyform, poll=.5,
                              timeout=5) == \
               components.rpcmessage_emptyform.as_dict

    def test_formquery_failure_dictd(self, crossbar, components):
        """
        Verify we raise proper exception for unexpected data types
        Only Form and Html objects should be allowed as inputs
        """
        with pytest.raises(TypeError):
            rpc_form_query('a1b2c3', dict(x=1, y='123'), poll=.5,
                           timeout=5)

    def test_rpc_send_failure_dictd(self, crossbar, components):
        """
        only Form and Html objects should be allowed as inputs
        """
        with pytest.raises(TypeError):
            rpc_send('a1b2c3', dict(x=1, y='123'))

    def test_formquery_timeout(self, crossbar, components):
        """
        mock RPC method won't return data until 2s delay.  Verify we get proper
        exception raise when our self-imposed timout is exceeded before
        receiving a response
        """
        with pytest.raises(RpcCallTimeout):
            assert rpc_form_query('a1b2c3', components.rpcform, poll=.5,
                                  timeout=1) == components.rpcmessage_form

    def test_formquery_timeout_disabled(self, crossbar, components):
        """
        rpc target will return False if there is no return value from the call.
        for browser.getformresponse this will mean no user has completed the
        last form sent to the UI yet.
        If have zero timeout, we will send and immediately check for response
        and no data will be available yet resulting in None as final return
        """
        assert rpc_form_query('a1b2c3', components.rpcform, poll=.5,
                              timeout=0) is False

    def test_formquery_timeout_forever(self, crossbar, components):
        """
        Block call forever and response should be returned at 2s delay
        """
        assert rpc_form_query('a1b2c3', components.rpcform, poll=.5,
                              timeout=-1) == components.rpcmessage_form.as_dict

    def test_formquery_invalid_target(self, crossbar, components):
        """
        If we call an invalid RPC target, verify proper exception is raised
        """
        with pytest.raises(ApplicationError):
            assert rpc_form_query('xyz', components.rpcform, poll=.5,
                                  timeout=2) == 'asdf'

    def test_rpc_displaymessage(self, crossbar, components):
        """
        When we send an RPC HTML message, the response back is "ok" from
        NetMagus backend.
        """
        assert rpc_send('a1b2c3', components.htmlform) == "ok"
        assert crossbar.userdata == components.rpcmessage_html.as_dict

    def test_invalid_ui_browser_method(self, crossbar, components):
        """
        Test case where object is sent to NM backend, but browser ui has not
        registered its RPC method yet and we receive exception
        """
        with pytest.raises(ApplicationError):
            assert rpc_send('d4e5f6', components.rpcform) == components \
                .rpcmessage_form.as_dict

    def test_formquery_slow_ack(self, crossbar, components):
        """
        test the case where NM backend doesn't ACK our transmission until 1
        second, causing rpc_send to retransmit
        """
        assert rpc_form_query('123456', components.rpcform, poll=.5,
                              timeout=-1) == components.rpcmessage_form.as_dict

    def test_output_json_examples(self, crossbar, components):
        print(json.dumps(components.rpcmessage_form.as_dict))
        print(json.dumps(components.rpcmessage_html.as_dict))



# coding=utf-8
from setuptools import find_packages, setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'future == 0.16.0',
    'autobahn-sync == 0.3.2'
]

test_requirements = [
    'pytest'
]

setup(name='netmagus',
      version='0.7.2',
      description='Python module for JSON data exchange via files or RPC with '
                  'the Intelligent Visibility NetMagus system.',
      long_description=readme + '\n\n' + history,
      url='http://www.intelligentvisibility.com/netmagus/',
      author='Richard Collins',
      author_email='richardc@intelligentvisibility.com',
      license='GNU lGPLv3',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: GNU Lesser General Public License v3 '
          'or later (LGPLv3+)',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Topic :: System :: Networking'
      ],
      keywords='netmagus network automation netops',
      packages=find_packages(include=['netmagus']),
      include_package_data=True,
      install_requires=requirements,
      test_suite='tests',
      tests_require=test_requirements
      )
